from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from phosphorus.lib.constants import licence_base_names

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


def get_license_files(base_dir: Path) -> Iterator[Path]:
    for base_name in licence_base_names:
        for file in chain(
            [base_dir.joinpath(base_name)],
            base_dir.glob(f"{base_name}.*"),
            base_dir.glob(f"{base_name}/**/*"),
        ):
            if file.is_file():
                yield file



from __future__ import annotations

from enum import IntEnum, unique
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable


@unique
class SGRParams(IntEnum):
    DEFAULT = 0
    BOLD = 1
    ITALIC = 3
    UNDERLINE = 4

    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    GREY = 37

    BLACK_BRIGHT = 90
    RED_BRIGHT = 91
    GREEN_BRIGHT = 92
    YELLOW_BRIGHT = 93
    BLUE_BRIGHT = 94
    MAGENTA_BRIGHT = 95
    CYAN_BRIGHT = 96
    WHITE_BRIGHT = 97

    @property
    def sequence(self) -> str:
        return f"\033[{self.value}m"


class SGRString(str):
    _sgr: str
    __slots__ = ("_sgr",)

    def __new__(  # noqa: PYI034
        cls, value: str, *, params: Iterable[SGRParams]
    ) -> SGRString:
        string = super().__new__(cls, value)
        prefix = "".join(param.sequence for param in params)
        suffix = SGRParams.DEFAULT.sequence
        string._sgr = f"{prefix}{value}{suffix}"  # noqa: SLF001
        return string

    def __str__(self) -> str:
        return self._sgr



from __future__ import annotations

import re
from dataclasses import dataclass, field

from phosphorus.lib.markers import Marker
from phosphorus.lib.packages import Package
from phosphorus.lib.regex import requirement_pattern
from phosphorus.lib.versions import Version, VersionClause


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Requirement:
    package: Package
    clauses: tuple[VersionClause, ...]
    marker: Marker

    @classmethod
    def from_string(cls, requirement: str) -> Requirement:  # TODO (py3.10): Use Self
        clause, *markers = requirement.split(";", maxsplit=1)
        marker = (
            Marker.from_string(markers[0].strip())
            if markers
            else Marker(boolean=None, markers=())
        )
        full_match = re.fullmatch(requirement_pattern, clause.strip())
        if not full_match:
            msg = "Could not parse the requirement"
            raise RuntimeError(msg)
        name = full_match["name"].strip()
        if full_match["clauses"]:
            clauses = tuple(
                VersionClause.from_string(matched_clause)
                for matched_clause in full_match["clauses"].strip(" ()").split(",")
            )
        else:
            clauses = ()
        return cls(package=Package(name=name), clauses=clauses, marker=marker)

    def __str__(self) -> str:
        name = self.package.name
        clauses = ",".join(str(clause) for clause in self.clauses)
        marker = str(self.marker)
        if clauses and marker:
            return f"{name} ({clauses}) ; {marker}"
        if clauses:
            return f"{name} ({clauses})"
        if marker:
            return f"{name} ; {marker}"
        return name


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class RequirementGroup:
    group: str
    requirements: tuple[Requirement, ...]


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class ResolvedRequirement:
    package: Package
    version: Version
    marker: Marker = field(default_factory=lambda: Marker(boolean=None, markers=()))

    @classmethod
    def from_string(
        cls, requirement: str
    ) -> ResolvedRequirement:  # TODO (py3.10): Use Self
        resolved_package, *markers = requirement.split(";", maxsplit=1)
        marker = (
            Marker.from_string(markers[0].strip())
            if markers
            else Marker(boolean=None, markers=())
        )
        name, matched_version = resolved_package.split("==")
        return cls(
            package=Package(name=name),
            version=Version.from_string(matched_version),
            marker=marker,
        )



from phosphorus.lib.parser import parse_args
from phosphorus.subcommands.build import BuildCommand
from phosphorus.subcommands.check import CheckCommand
from phosphorus.subcommands.install import InstallCommand
from phosphorus.subcommands.lock import LockCommand


def main() -> None:
    args = parse_args()
    if args.subcommand == "build":  # TODO (py3.9): Use match
        BuildCommand(args).run()
    elif args.subcommand == "check":
        CheckCommand(args).run()
    elif args.subcommand == "install":
        InstallCommand(args).run()
    elif args.subcommand == "lock":
        LockCommand(args).run()



from __future__ import annotations

import sys

MINOR = sys.version_info.minor


if MINOR >= 9:  # noqa: PLR2004
    remove_prefix = str.removeprefix
else:

    def remove_prefix(s: str, prefix: str) -> str:  # TODO (py3.9): Use str.removeprefix
        if s.startswith(prefix):
            return s[len(prefix) :]
        return s


if MINOR >= 11:  # noqa: PLR2004
    import tomllib

    toml_parser = tomllib.load
    TOMLDecodeError = tomllib.TOMLDecodeError
else:  # TODO (py3.10): Use tomllib
    import tomli

    toml_parser = tomli.load
    TOMLDecodeError = tomli.TOMLDecodeError



from __future__ import annotations

import json
import os
from dataclasses import asdict, dataclass
from hashlib import sha256
from importlib.machinery import SourceFileLoader
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from trove_classifiers import classifiers

from phosphorus._seven import toml_parser
from phosphorus.lib.constants import (
    BooleanOperator,
    ComparisonOperator,
    MarkerVariable,
    lock_file_name,
    pyproject_base_name,
)
from phosphorus.lib.contributors import Contributor
from phosphorus.lib.exceptions import (
    ImproperlyConfiguredProjectError,
    MissingProjectRootError,
)
from phosphorus.lib.packages import Package
from phosphorus.lib.requirements import Requirement, RequirementGroup
from phosphorus.lib.tags import Tag
from phosphorus.lib.versions import Version, VersionClause

if TYPE_CHECKING:
    from collections.abc import Iterable


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Script:
    command: str
    entrypoint: str


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class ProjectURL:
    name: str
    url: str


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class LocalPackage:
    base_dir: Path
    path: Path

    @property
    def absolute_path(self) -> Path:
        return self.base_dir.joinpath(self.path)


class JSONEncoder(json.JSONEncoder):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.base_dir = kwargs.pop("base_dir")
        super().__init__(*args, **kwargs)

    def default(self, o: Any) -> Any:
        if isinstance(o, Path):  # TODO (py3.9): Use match
            if o == Path(os.devnull):
                return os.devnull
            if o.is_absolute():
                return o.relative_to(self.base_dir).as_posix()
            return o.as_posix()
        if isinstance(o, ComparisonOperator):
            return str(o)
        if isinstance(o, BooleanOperator):
            return str(o)
        if isinstance(o, MarkerVariable):
            return str(o)

        return super().default(o)


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Metadata:
    base_dir: Path
    package: Package
    version: Version
    summary: str
    homepage: str
    license: str
    readme: Path
    keywords: tuple[str, ...]
    tags: tuple[Tag, ...]
    authors: tuple[Contributor, ...]
    maintainers: tuple[Contributor, ...]
    python: tuple[VersionClause, ...]
    classifiers: tuple[str, ...]
    requirement_groups: tuple[RequirementGroup, ...]
    scripts: tuple[Script, ...]
    project_urls: tuple[ProjectURL, ...]
    package_paths: tuple[LocalPackage, ...]

    @classmethod
    def from_path(cls, path: Path | None = None) -> Metadata:  # TODO (py3.10): Use Self
        settings_path = get_pyproject(path)
        settings = get_settings(settings_path)

        urls = settings.get("urls", {})
        python = settings["requires-python"]

        return cls(
            base_dir=settings_path.parent,
            package=get_package(settings),
            version=get_version(settings),
            summary=settings.get("description", ""),
            homepage=urls.get("homepage", ""),
            license=get_license(settings),
            readme=settings_path.parent.joinpath(settings.get("readme", os.devnull)),
            keywords=keep_unique(settings.get("keywords", [])),
            tags=(Tag(interpreter="py3", abi=None, platform="any"),),
            authors=get_contributors(settings.get("authors", [])),
            maintainers=get_contributors(settings.get("maintainers", [])),
            python=keep_unique(
                VersionClause.from_string(clause) for clause in python.split(",")
            ),
            classifiers=get_classifiers(settings.get("classifiers", [])),
            requirement_groups=get_requirements(
                settings.get("dependencies", []), settings.get("dev_dependencies", {})
            ),
            project_urls=keep_unique(
                ProjectURL(name=name, url=url)
                for name, url in urls.items()
                if name != "homepage"
            ),
            scripts=keep_unique(
                Script(command=command, entrypoint=entrypoint)
                for command, entrypoint in settings.get("scripts", {}).items()
            ),
            package_paths=get_package_paths(settings, settings_path.parent),
        )

    @property
    def lockfile(self) -> Path:
        return self.base_dir.joinpath(lock_file_name)

    @property
    def pyproject(self) -> Path:
        return self.base_dir.joinpath(pyproject_base_name)

    @property
    def p_hash(self) -> str:
        json_dump = json.dumps(
            asdict(self), cls=JSONEncoder, base_dir=self.base_dir, sort_keys=True
        )
        return sha256(json_dump.encode("utf-8")).hexdigest()


def keep_unique(items: Iterable[Any]) -> tuple[Any, ...]:
    return tuple(sorted(set(items)))


def get_pyproject(cwd: Path | None) -> Path:
    if cwd is None:
        cwd = Path.cwd().resolve()

    while not (pyproject := cwd.joinpath(pyproject_base_name)).is_file():
        if cwd.as_posix() == "/":
            raise MissingProjectRootError(cwd)
        cwd = cwd.parent
    return pyproject


def get_contributors(settings: list[dict[str, str]]) -> tuple[Contributor, ...]:
    return tuple(Contributor.from_data(data) for data in settings)


def get_settings(settings_path: Path) -> dict[str, Any]:
    with settings_path.open("rb") as settings_file:
        all_settings = toml_parser(settings_file)
    settings = all_settings.get("project", {})
    phosphorus_settings = all_settings.get("tool", {}).get("phosphorus", {})
    settings["dev_dependencies"] = phosphorus_settings.get("dev-dependencies", {})
    settings["dynamic_definitions"] = phosphorus_settings.get("dynamic", {})
    settings["included_packages"] = phosphorus_settings.get("packages", {})
    return cast("dict[str, Any]", settings)  # TODO (py3.9): Use dict[str, Any]


def get_package(settings: dict[str, Any]) -> Package:
    return Package(name=settings["name"])


def get_version(settings: dict[str, Any]) -> Version:
    version_key = "version"
    version = settings.get(version_key)
    if version:
        return Version.from_string(version)
    try:
        version = settings["dynamic_definitions"][version_key]["file"]
    except KeyError as exc:
        raise ImproperlyConfiguredProjectError(version_key) from exc
    name = "_module"
    _module = SourceFileLoader(name, version).load_module(name)
    version = _module.__version__
    return Version.from_string(version)


def get_license(settings: dict[str, Any]) -> str:
    license_info = settings.get("license", {})
    if license_text := license_info.get("text", ""):
        return cast(str, license_text)
    return ""


def get_classifiers(user_classifiers: list[str]) -> tuple[str, ...]:
    unique_classifiers = keep_unique(user_classifiers)
    if any(classifier not in classifiers for classifier in unique_classifiers):
        classifier_key = "classifiers"
        raise ImproperlyConfiguredProjectError(classifier_key)
    return unique_classifiers


def get_requirements(
    dependencies: list[str], dev_dependencies: dict[str, list[str]]
) -> tuple[RequirementGroup, ...]:
    output = {
        group: [Requirement.from_string(constraint) for constraint in requirement_info]
        for group, requirement_info in dev_dependencies.items()
    }
    output["main"] = [
        Requirement.from_string(constraint) for constraint in dependencies
    ]

    return keep_unique(
        RequirementGroup(group=group, requirements=keep_unique(requirements))
        for group, requirements in output.items()
    )


def get_package_paths(
    settings: dict[str, Any], base_dir: Path
) -> tuple[LocalPackage, ...]:
    package_key = "included_packages"
    if packages := settings.get(package_key, []):
        output = packages
    elif base_dir.joinpath("src").exists():
        output = ["src"]
    else:
        raise ImproperlyConfiguredProjectError(package_key)
    return keep_unique(
        LocalPackage(path=Path(path), base_dir=base_dir) for path in output
    )



from __future__ import annotations

import json
from http import HTTPStatus
from threading import Thread
from typing import TYPE_CHECKING
from urllib.error import HTTPError
from urllib.request import Request, urlopen

from phosphorus.lib.constants import pypi_cache

if TYPE_CHECKING:
    from phosphorus.lib.packages import Package, VersionedPackage, VersionInfo
    from phosphorus.lib.versions import Version


def get_version_info(
    packages: list[VersionedPackage], *, last_check: float = float("inf")
) -> dict[VersionedPackage, VersionInfo]:

    threads: list[Thread] = []
    results: dict[VersionedPackage, VersionInfo] = {}
    for package_info in packages:
        thread = Thread(
            target=_get_version_info,
            args=(package_info.package, package_info.version, last_check, results),
        )
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    return results


def _get_version_info(
    package: Package,
    version: Version | None,
    last_check: float,
    results: dict[VersionedPackage, VersionInfo],
) -> None:
    from phosphorus.lib.packages import VersionedPackage, VersionInfo

    key = VersionedPackage(package, version)
    filename = "latest.json" if version is None else f"{version.canonical_form}.json"
    local_cache = pypi_cache.joinpath(package.name, filename)
    if local_cache.exists():
        local_info = VersionInfo.from_cache(local_cache)
        modified = local_cache.stat().st_mtime
        if modified >= last_check:
            results[key] = local_info
            return
        etag = local_info.etag
    else:
        local_cache.parent.mkdir(parents=True, exist_ok=True)
        etag = ""

    try:
        pypi_info = _get_info_from_pypi(package, version, etag)
    except HTTPError as exc:
        if exc.status == HTTPStatus.NOT_MODIFIED:
            local_cache.touch()
            results[key] = local_info
            return
        raise

    with local_cache.open("w") as file:
        json.dump(pypi_info.as_dict(), file, indent=4)
    results[key] = pypi_info


def _get_info_from_pypi(
    package: Package, version: Version | None, etag: str
) -> VersionInfo:
    from phosphorus.lib.packages import VersionInfo

    package_slug = f"{package.name}/{version}" if version else package.name
    url = f"https://pypi.org/pypi/{package_slug}/json"
    headers = {}
    if etag:
        headers["If-None-Match"] = etag
    request = Request(url, headers=headers)  # noqa: S310
    try:
        response = urlopen(request, timeout=60)  # noqa: S310
    except TimeoutError as exc:
        msg = "pypi connection timed out"
        raise RuntimeError(msg) from exc
    return VersionInfo.from_response(response)



from __future__ import annotations

import os
from pathlib import Path
from typing import Any

from phosphorus.construction.sdist import SdistBuilder
from phosphorus.construction.wheel import WheelBuilder

# mandatory hooks


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    metadata_path = None if metadata_directory is None else Path(metadata_directory)
    builder = WheelBuilder(Path(wheel_directory), config_settings, metadata_path)
    return builder.build().name


def build_sdist(
    sdist_directory: str, config_settings: dict[str, Any] | None = None
) -> str:
    builder = SdistBuilder(Path(sdist_directory), config_settings, None)
    return builder.build().name


# optional hooks


def prepare_metadata_for_build_wheel(
    metadata_directory: str, config_settings: dict[str, Any] | None = None
) -> str:
    builder = WheelBuilder(Path(os.devnull), config_settings, Path(metadata_directory))
    return builder.prepare_metadata().name


def build_editable(
    wheel_directory: str,
    config_settings: dict[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    metadata_path = None if metadata_directory is None else Path(metadata_directory)
    builder = WheelBuilder(
        Path(wheel_directory), config_settings, metadata_path, editable=True
    )
    return builder.build().name



from __future__ import annotations

import hashlib
from base64 import urlsafe_b64encode
from dataclasses import dataclass
from stat import S_ISDIR
from tarfile import TarInfo
from typing import TYPE_CHECKING
from zipfile import ZipInfo

if TYPE_CHECKING:
    from pathlib import Path


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class ArchiveFile:
    absolute_path: Path
    base_dir: Path
    digest: str
    size: int
    mode: int

    @classmethod
    def from_file(
        cls, source: Path, base_dir: Path
    ) -> ArchiveFile:  # TODO (py3.10): Use Self
        stat = source.stat()

        return cls(
            absolute_path=source,
            base_dir=base_dir,
            digest=cls.hash_file(source),
            size=stat.st_size,
            mode=stat.st_mode,
        )

    @property
    def relative_path(self) -> Path:
        return self.absolute_path.relative_to(self.base_dir)

    @property
    def zip_info(self) -> ZipInfo:
        path = self.relative_path
        date_time = (1980, 1, 1, 0, 0, 0)
        zip_info = ZipInfo(path.as_posix(), date_time=date_time)
        zip_info.external_attr = self.normalised_mode()
        return zip_info

    @property
    def tar_info(self) -> TarInfo:
        path = self.relative_path
        tarinfo = TarInfo(path.as_posix())
        tarinfo.mtime = 0
        tarinfo.uid = 0
        tarinfo.gid = 0
        tarinfo.uname = ""
        tarinfo.gname = ""
        tarinfo.mode = self.normalised_mode(for_zip=False)
        tarinfo.size = self.size
        return tarinfo

    def normalised_mode(self, *, for_zip: bool = True) -> int:
        mode = (self.mode | 0o644) & ~0o133
        if self.mode & 0o100:
            mode |= 0o111
        if for_zip:
            mode = (mode & 0xFFFF) << 16
            if S_ISDIR(self.mode):
                mode |= 0x10

        return mode

    @staticmethod
    def hash_file(path: Path, buffer_size: int = 2**16) -> str:
        sha256 = hashlib.sha256()

        with path.open("rb") as f:
            while data := f.read(buffer_size):
                sha256.update(data)

        hash_value = urlsafe_b64encode(sha256.digest()).decode("ascii").rstrip("=")
        return f"sha256={hash_value}"



from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace

from phosphorus.__version__ import __version__

sys.tracebacklimit = 0


def add_boolean_optional_action(
    parser: ArgumentParser,
    var: str,
    help_text: str = "",
    *,
    required: bool = False,
    default: bool | None = None,
) -> None:  # TODO (py3.8): Use BooleanOptionalAction
    dest = var.replace("-", "_")
    parser.add_argument(
        f"--{var}",
        action="store_true",
        dest=dest,
        default=default,
        help=help_text,
        required=required,
    )
    parser.add_argument(
        f"--no-{var}",
        action="store_false",
        dest=dest,
        default=default,
        help=help_text,
        required=required,
    )


def parse_args() -> Namespace:
    parser = ArgumentParser(description="a dependency management tool")
    parser = ArgumentParser(description="git repository manager")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"phosphorus {__version__}",
        help="print the version and exit",
    )

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        dest="verbosity",
        help="increase the level of verbosity",
    )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    build_parser = subparsers.add_parser(
        "build",
        parents=[parent_parser],
        help="build the wheel and the sdist distributions for the package",
    )
    add_boolean_optional_action(
        build_parser,
        "sdist",
        default=True,
        help_text="build the sdist distribution",
    )
    add_boolean_optional_action(
        build_parser,
        "wheel",
        default=True,
        help_text="build the wheel distribution",
    )

    check_parser = subparsers.add_parser(
        "check", parents=[parent_parser], help="check that the lock file is up to date"
    )
    add_boolean_optional_action(
        check_parser,
        "lockfile",
        default=False,
        help_text="check if the lock file is up to date",
    )
    add_boolean_optional_action(
        check_parser,
        "outdated",
        default=False,
        help_text="check if the dependencies are outdated",
    )

    install_parser = subparsers.add_parser(
        "install", parents=[parent_parser], help="install the project dependencies"
    )
    add_boolean_optional_action(
        install_parser,
        "sync",
        default=False,
        help_text="sync the dependencies, by removing the ones that are not in the lock file",
    )
    dependency_groups = install_parser.add_mutually_exclusive_group()
    dependency_groups.add_argument(
        "-e",
        "--exclude",
        action="append",
        default=[],
        help="exclude a group of dependencies",
    )
    dependency_groups.add_argument(
        "-g",
        "--groups",
        action="append",
        default=[],
        help="install a group of dependencies",
    )

    lock_parser = subparsers.add_parser(
        "lock", parents=[parent_parser], help="lock the project dependencies"
    )
    lock_parser.add_argument(
        "-f", "--force", action="store_true", help="force recreating the lock"
    )
    add_boolean_optional_action(
        lock_parser,
        "enforce-pep440",
        default=True,
        help_text="enforce PEP 440",
    )
    add_boolean_optional_action(
        lock_parser,
        "allow-pre-releases",
        default=False,
        help_text="allow pre-releases",
    )
    add_boolean_optional_action(
        lock_parser,
        "allow-dev-releases",
        default=False,
        help_text="allow dev releases",
    )

    args = parser.parse_args()
    if args.verbosity > 0:
        sys.tracebacklimit = 1000

    return args



from __future__ import annotations

from itertools import chain
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any

from phosphorus.lib.contributors import Contributor
from phosphorus.lib.metadata import Metadata

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from phosphorus.lib.zipped_file import ArchiveFile


class Builder:
    __slots__ = ("config", "metadata_dir", "output_dir", "meta")

    def __init__(
        self, output_dir: Path, config: dict[str, Any] | None, metadata_dir: Path | None
    ) -> None:
        self.output_dir = output_dir
        self.config = config or {}
        self.metadata_dir = metadata_dir
        self.meta = Metadata.from_path()

    def build(self) -> Path:
        package = self.output_dir.joinpath(self.filename)
        self.output_dir.mkdir(parents=True, exist_ok=True)
        package.unlink(missing_ok=True)

        with TemporaryDirectory() as temp_dir_name:
            temp_dir = Path(temp_dir_name).resolve()
            files = chain(
                self.package_files(temp_dir), self.non_package_files(temp_dir)
            )
            self.write_files(files, package, temp_dir)

        return package

    @property
    def filename(self) -> str:
        raise NotImplementedError

    def package_files(self, temp_dir: Path) -> Iterator[ArchiveFile]:
        raise NotImplementedError

    def non_package_files(self, temp_dir: Path) -> Iterator[ArchiveFile]:
        raise NotImplementedError

    def get_info_file(self, temp_dir: Path, data: Any) -> ArchiveFile:
        raise NotImplementedError

    def write_files(
        self, files: Iterable[ArchiveFile], package: Path, temp_dir: Path
    ) -> None:
        raise NotImplementedError

    @property
    def base_name(self) -> str:
        package_name = self.meta.package.distribution_name
        return f"{package_name}-{self.meta.version}"

    def get_metadata_content(self) -> Iterator[str]:
        yield "Metadata-Version: 2.3"
        yield f"Name: {self.meta.package.name}"
        yield f"Version: {self.meta.version}"
        if self.meta.summary:
            yield f"Summary: {self.meta.summary}"

        if self.meta.homepage:
            yield f"Home-page: {self.meta.homepage}"

        if self.meta.license:
            yield f"License: {self.meta.license}"

        if self.meta.keywords:
            keyword_string = ",".join(self.meta.keywords)
            yield f"Keywords: {keyword_string}"

        if names := Contributor.stringify_names(self.meta.authors):
            yield f"Author: {names}"

        if emails := Contributor.stringify_emails(self.meta.authors):
            yield f"Author-email: {emails}"

        if names := Contributor.stringify_names(self.meta.maintainers):
            yield f"Maintainer: {names}"

        if emails := Contributor.stringify_emails(self.meta.maintainers):
            yield f"Maintainer-email: {emails}"

        python_versions = ",".join(str(clause) for clause in self.meta.python)
        yield f"Requires-Python: {python_versions}"

        for classifier in self.meta.classifiers:
            yield f"Classifier: {classifier}"

        for requirement_group in self.meta.requirement_groups:
            if requirement_group.group == "main":
                for requirement in requirement_group.requirements:
                    yield f"Requires-Dist: {requirement}"
                break

        for project_url in self.meta.project_urls:
            yield f"Project-URL: {project_url.name.title()}, {project_url.url}"

        if readme_text := self.meta.readme.read_text():
            yield "Description-Content-Type: text/markdown"
            yield ""
            yield readme_text.rstrip("\n")



from __future__ import annotations

from typing import TYPE_CHECKING

from phosphorus.construction.api import build_sdist, build_wheel
from phosphorus.lib.term import SGRParams, SGRString
from phosphorus.subcommands.base import BaseCommand

if TYPE_CHECKING:
    from argparse import Namespace


class BuildCommand(BaseCommand):
    __slots__ = ("build_sdist", "build_wheel")

    def __init__(self, args: Namespace, /) -> None:
        super().__init__(args)
        self.build_sdist = args.sdist
        self.build_wheel = args.wheel

    def run(self) -> None:
        package_name = SGRString(self.meta.package.name, params=[SGRParams.CYAN])
        version = SGRString(f"({self.meta.version})", params=[SGRParams.BOLD])
        print(f"Building {package_name} {version} ...")
        dist_dir = self.meta.base_dir.joinpath("dist")
        dist_dir.mkdir(exist_ok=True)
        if self.build_sdist:
            print(f"🔧 Building {SGRString('sdist', params=[SGRParams.MAGENTA])} ...")
            sdist = build_sdist(dist_dir.as_posix())
            sgr_sdist = SGRString(str(sdist), params=[SGRParams.BLUE])
            print(f"✅ {sgr_sdist} built successfully!")
        if self.build_wheel:
            print(f"🔧 Building {SGRString('wheel', params=[SGRParams.MAGENTA])} ...")
            wheel = build_wheel(dist_dir.as_posix())
            sgr_wheel = SGRString(str(wheel), params=[SGRParams.BLUE])
            print(f"✅ {sgr_wheel} built successfully!")



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class Tag:
    interpreter: str
    abi: str | None
    platform: str

    def __str__(self) -> str:
        abi = self.abi or "none"
        return f"{self.interpreter}-{abi}-{self.platform}"



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Contributor:
    name: str
    email: str

    @classmethod
    def from_data(cls, data: dict[str, str]) -> Contributor:  # TODO (py3.10): Use Self
        return cls(name=data.get("name", ""), email=data.get("email", ""))

    def formatted_email(self) -> str:
        if not self.email:
            return ""

        if not self.name:
            return self.email

        return f'"{self.name}" <{self.email}>'

    @staticmethod
    def stringify_names(contributors: Iterable[Contributor]) -> str:
        return ",".join(
            contributor.name for contributor in contributors if contributor.name
        )

    @staticmethod
    def stringify_emails(contributors: Iterable[Contributor]) -> str:
        return ",".join(
            contributor.formatted_email()
            for contributor in contributors
            if contributor.email
        )



import re

raw_version_pattern = r"""
    ^
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
    $
"""
raw_requirement_pattern = r"""
    (?P<name>[a-z0-9\[\]]+(?:[a-z0-9-_.])*)+               # package name
    (?P<clauses>.*)                                        # comma-separated list of clauses
"""
version_pattern = re.compile(raw_version_pattern, re.VERBOSE | re.IGNORECASE)
requirement_pattern = re.compile(raw_requirement_pattern, re.VERBOSE | re.IGNORECASE)
version_separators = re.compile(r"[._-]")



from __future__ import annotations

from dataclasses import dataclass
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from phosphorus._seven import remove_prefix
from phosphorus.lib.constants import hash_prefix
from phosphorus.lib.exceptions import ThirdPartyError
from phosphorus.lib.packages import Package
from phosphorus.lib.requirements import ResolvedRequirement
from phosphorus.lib.uv_runner import uv_run

if TYPE_CHECKING:
    from collections.abc import Iterator

    from phosphorus.lib.requirements import RequirementGroup
    from phosphorus.lib.types import ResolvedPackageInfo


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class LockEntry:
    requirement: ResolvedRequirement
    groups: tuple[str, ...]
    hashes: tuple[str, ...]


class Resolver:
    __slots__ = (
        "requirement_groups",
        "enforce_pep440",
        "allow_pre_releases",
        "allow_dev_releases",
        "packages",
        "resolved_packages",
        "verbosity",
    )

    def __init__(
        self,
        requirement_groups: tuple[RequirementGroup, ...],
        *,
        verbosity: int,
        enforce_pep440: bool,
        allow_pre_releases: bool,
        allow_dev_releases: bool,
    ) -> None:
        self.requirement_groups = requirement_groups
        self.enforce_pep440 = enforce_pep440
        self.allow_pre_releases = allow_pre_releases
        self.allow_dev_releases = allow_dev_releases
        self.verbosity = verbosity
        self.packages = {
            requirement.package: requirement_group.group
            for requirement_group in self.requirement_groups
            for requirement in requirement_group.requirements
        }
        self.resolved_packages: dict[Package, ResolvedPackageInfo] = {}

    def resolve(self) -> list[LockEntry]:
        with NamedTemporaryFile("w", delete=False) as tmp:
            for group in self.requirement_groups:
                for requirement in group.requirements:
                    tmp.write(f"{requirement}\n")
            tmp.seek(0)
            output = uv_run(
                ["pip", "compile", "--generate-hashes", "--universal", tmp.name]
            )
            if output.returncode:
                if self.verbosity:
                    print(output.stderr.decode())
                command = "uv pip compile"
                raise ThirdPartyError(command)
            print("✅ Done!")

        for raw_package_info in self.split_resolution(output.stdout.decode()):
            self.preprocess(raw_package_info)

        return [
            self.get_lock_entry(resolved_package)
            for resolved_package in self.resolved_packages.values()
        ]

    @staticmethod
    def split_resolution(resolution: str) -> Iterator[list[str]]:
        lines: list[str] = []
        should_yield = False
        for line in resolution.splitlines():
            stripped = line.strip(" \\")
            if not stripped:
                pass
            elif stripped.startswith("#"):
                lines.append(stripped)
                should_yield = True
            else:
                if should_yield:
                    if lines:
                        if any(not line.startswith("#") for line in lines):
                            yield lines
                        lines = []
                    should_yield = False
                lines.append(stripped)
        if lines and any(not line.startswith("#") for line in lines):
            yield lines

    def preprocess(self, raw_package_info: list[str]) -> None:
        requirement = ResolvedRequirement.from_string(raw_package_info[0])
        parents = {
            Package(name=stripped)
            for line in raw_package_info
            if line.startswith("#")
            and (stripped := remove_prefix(line, "# via").lstrip(" #"))
            and not stripped.startswith("-r")
        }

        self.resolved_packages[requirement.package] = {
            "requirement": requirement,
            "hashes": tuple(
                file_hash[len(hash_prefix) :]
                for file_hash in raw_package_info[1:]
                if file_hash.startswith(hash_prefix)
            ),
            "parents": parents,
        }

    def get_lock_entry(self, resolved_package: ResolvedPackageInfo) -> LockEntry:
        group = self.packages.get(resolved_package["requirement"].package, "")
        groups = {group} if group else set()
        parents = resolved_package["parents"].copy()
        seen = set()
        while parents:
            parent = parents.pop()
            if parent in seen:
                continue
            seen.add(parent)
            if group := self.packages.get(parent, ""):
                groups.add(group)
                continue
            parents.update(self.resolved_packages[parent]["parents"])
        return LockEntry(
            requirement=resolved_package["requirement"],
            groups=tuple(sorted(groups)),
            hashes=resolved_package["hashes"],
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, TypedDict

if TYPE_CHECKING:
    from phosphorus.lib.packages import Package
    from phosphorus.lib.requirements import ResolvedRequirement
    from phosphorus.lib.resolver import LockEntry
    from phosphorus.lib.versions import Version

Match = Optional[str]  # TODO (py3.9): Use |


class ResolvedPackageInfo(TypedDict):
    requirement: ResolvedRequirement
    hashes: tuple[str, ...]
    parents: set[Package]


class PackageDiff(TypedDict):
    update: set[tuple[ResolvedRequirement | None, LockEntry]]
    remove: set[ResolvedRequirement]


class InstalledVersions(TypedDict, total=False):
    package: Package
    installed_version: Version
    latest_version: Version



from __future__ import annotations

import json
import sys
from typing import TYPE_CHECKING

from phosphorus._seven import TOMLDecodeError
from phosphorus.lib.packages import Package, VersionedPackage
from phosphorus.lib.pypi import get_version_info
from phosphorus.lib.uv_runner import uv_run
from phosphorus.lib.versions import Version
from phosphorus.subcommands.base import BaseCommand

if TYPE_CHECKING:
    from argparse import Namespace

    from phosphorus.lib.types import InstalledVersions


class CheckCommand(BaseCommand):
    __slots__ = ("outdated", "lockfile")

    def __init__(self, _args: Namespace, /) -> None:
        super().__init__(_args)
        self.lockfile = _args.lockfile
        self.outdated = _args.outdated

    def run(self) -> None:
        if not self.outdated and not self.lockfile:
            print("❌ No check selected", file=sys.stderr)
            sys.exit(2)
        if self.outdated:
            self.check_outdated()
        if self.lockfile:
            try:
                self.check_lockfile()
            except RuntimeError as exc:
                print(f"❌ Lockfile check failed: {exc}", file=sys.stderr)
                sys.exit(1)
            else:
                print(
                    "✅ Success: Lockfile is valid and up-to-date with pyproject.toml"
                )

    def check_lockfile(self) -> None:
        try:
            p_hash = self.get_current_hash()
        except FileNotFoundError as exc:
            msg = "No lockfile found"
            raise RuntimeError(msg) from exc
        except (KeyError, TOMLDecodeError) as exc:
            msg = "Lockfile is invalid"
            raise RuntimeError(msg) from exc

        if p_hash != self.meta.p_hash:
            msg = "Hashes don't match"
            raise RuntimeError(msg)

    def check_outdated(self) -> None:
        pip_list = uv_run(["pip", "list", "--exclude-editable", "--format=json"])
        installed_packages: list[InstalledVersions] = [
            {
                "package": Package(package_info["name"]),
                "installed_version": Version.from_string(package_info["version"]),
            }
            for package_info in json.loads(pip_list.stdout)
        ]
        version_info = get_version_info(
            packages=[
                VersionedPackage(package_info["package"])
                for package_info in installed_packages
            ],
        )
        outdated: list[InstalledVersions] = []
        for package_info in installed_packages:
            versioned_package = VersionedPackage(package_info["package"])
            latest_version = version_info[versioned_package].version
            installed_version = package_info["installed_version"]
            if latest_version != installed_version:
                outdated.append(
                    {
                        "package": package_info["package"],
                        "installed_version": installed_version,
                        "latest_version": latest_version,
                    }
                )
        if outdated:
            self.print_outdated(outdated)
        else:
            print("✅ All installed packages are up-to-date!")

    @staticmethod
    def print_outdated(outdated: list[InstalledVersions]) -> None:
        header = ("Package", "Installed version", "Latest version")
        max_package = len(header[0])
        max_installed = len(header[1])
        max_latest = len(header[2])

        rows = []
        for outdated_package in reversed(outdated):
            package = outdated_package["package"].name
            max_package = max(max_package, len(package))
            installed_version = str(outdated_package["installed_version"])
            max_installed = max(max_installed, len(installed_version))
            latest_version = str(outdated_package["latest_version"])
            max_latest = max(max_latest, len(latest_version))
            rows.append((package, installed_version, latest_version))

        print(
            header[0].center(max_package),
            " ",
            header[1].center(max_installed),
            " ",
            header[2].center(max_latest),
        )
        print(
            "-" * max_package,
            " ",
            "-" * max_installed,
            " ",
            "-" * max_latest,
        )
        for row in reversed(rows):
            print(
                row[0].ljust(max_package),
                " ",
                row[1].rjust(max_installed),
                " ",
                row[2].rjust(max_latest),
            )



from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from phosphorus.lib.pypi import get_version_info
from phosphorus.lib.utils import canonicalise_name
from phosphorus.lib.versions import Version, VersionClause

if TYPE_CHECKING:
    from collections.abc import Iterable
    from http.client import HTTPResponse
    from pathlib import Path

    from phosphorus.lib.requirements import Requirement


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class VersionedPackage:
    """
    A package with an optional version.

    When the version is None, it represents the latest version.
    """

    package: Package
    version: Version | None = None


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class ResolvedPackage:
    package: Package
    version: Version
    yanked: bool
    requires_python: list[VersionClause]
    requires_dist: list[Requirement]


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class VersionInfo:
    etag: str
    requires_dist: list[str]
    requires_python: str
    yanked: bool
    package: Package
    version: Version
    releases: list[Version] | None = None

    @classmethod
    def from_cache(cls, cache_path: Path) -> VersionInfo:  # TODO (py3.10): Use Self
        with cache_path.open() as file:
            info: dict[str, Any] = json.load(file)
        releases = (
            [Version.from_string(release) for release in info["releases"]]
            if "releases" in info
            else None
        )
        return cls(
            etag=info["ETag"],
            requires_dist=info["requires_dist"],
            requires_python=info["requires_python"],
            yanked=info["yanked"],
            package=Package(cache_path.parent.name),
            version=Version.from_string(info["version"]),
            releases=releases,
        )

    @classmethod
    def from_response(
        cls, response: HTTPResponse
    ) -> VersionInfo:  # TODO (py3.10): Use Self
        info = json.loads(response.read())
        releases = (
            [Version.from_string(release) for release in info["releases"]]
            if "releases" in info
            else None
        )
        return cls(
            etag=response.headers["ETag"],
            requires_dist=info["info"]["requires_dist"],
            requires_python=info["info"]["requires_python"],
            yanked=info["info"]["yanked"],
            package=Package(info["info"]["name"]),
            version=Version.from_string(info["info"]["version"]),
            releases=releases,
        )

    def as_dict(self) -> dict[str, Any]:
        out = {
            "ETag": self.etag,
            "requires_dist": self.requires_dist,
            "requires_python": self.requires_python,
            "yanked": self.yanked,
            "package": self.package.name,
            "version": str(self.version),
        }
        if self.releases is not None:
            releases = [str(release) for release in self.releases]
            out["releases"] = releases
        return out


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Package:
    name: str
    distribution_name: str = field(repr=False, compare=False)

    def __init__(self, name: str) -> None:
        name = canonicalise_name(name)
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "distribution_name", name.replace("-", "_"))

    def __str__(self) -> str:
        return self.name

    def get_versions(
        self,
        *,
        enforce_pep440: bool = False,
        allow_pre_releases: bool = True,
        allow_dev_releases: bool = True,
        clauses: Iterable[VersionClause] = (),
        last_check: float = float("inf"),
    ) -> list[Version]:

        latest_package = VersionedPackage(package=self)
        version_info = get_version_info(
            packages=[latest_package], last_check=last_check
        )[latest_package]
        versions: list[Version] = []
        if version_info.releases is None:
            msg = "No releases found"
            raise RuntimeError(msg)
        for version in version_info.releases:
            if version.is_pre_release and not allow_pre_releases:
                continue
            if version.is_dev_release and not allow_dev_releases:
                continue
            if not version.pep_440_compliant and enforce_pep440:
                continue
            for clause in clauses:
                if not clause.match(version):
                    break
            else:
                versions.append(version)
        return sorted(versions)



from __future__ import annotations

import ast
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from phosphorus.lib.constants import (
    BooleanOperator,
    ComparisonOperator,
    MarkerVariable,
    TokenRule,
)
from phosphorus.lib.exceptions import UnreachableCodeError
from phosphorus.lib.utils import canonicalise_name
from phosphorus.lib.versions import Version, VersionClause

if TYPE_CHECKING:
    import re
    from collections.abc import Iterator


@dataclass(frozen=True)  # TODO (py3.9): Use slots=True
class MarkerAtom:
    variable: MarkerVariable
    operator: ComparisonOperator
    value: str

    def __str__(self) -> str:
        return f"{self.variable} {self.operator} '{self.value}'"

    @classmethod
    def from_dict(cls, specs: dict[str, Any]) -> MarkerAtom:  # TODO (py3.10): Use Self
        variable = specs["variable"]
        if isinstance(variable, str):
            variable = MarkerVariable(variable)

        operator = specs["operator"]
        if isinstance(operator, str):
            operator = ComparisonOperator(operator)

        return cls(variable=variable, operator=operator, value=specs["value"])

    def evaluate(self, extra: str = "") -> bool:
        environment_value = self.variable.get_env_value(extra)
        if self.operator == ComparisonOperator.IN:  # TODO (py3.9): Use match
            return environment_value in self.value
        if self.operator == ComparisonOperator.NOT_IN:
            return environment_value in self.value
        if self.variable in {
            MarkerVariable.PYTHON_VERSION,
            MarkerVariable.PYTHON_FULL_VERSION,
            MarkerVariable.IMPLEMENTATION_VERSION,
            MarkerVariable.PLATFORM_VERSION,
        }:
            clause = VersionClause(
                operator=self.operator,
                identifier=Version.from_string(self.value),
            )
            return clause.match(Version.from_string(environment_value))
        if self.operator in {
            ComparisonOperator.COMPATIBLE_WITH,
            ComparisonOperator.EXACT_MATCH,
        }:
            msg = f"Cannot compare `{self.variable}` with `{self.operator}`"
            raise ValueError(msg)
        if self.operator == ComparisonOperator.EQUAL_TO:
            return environment_value == self.value
        if self.operator == ComparisonOperator.NOT_EQUAL:
            return environment_value != self.value
        if self.operator == ComparisonOperator.LESS_OR_EQUAL:
            return environment_value <= self.value
        if self.operator == ComparisonOperator.GREATER_OR_EQUAL:
            return environment_value >= self.value
        if self.operator == ComparisonOperator.LESS_THAN:
            return environment_value < self.value
        if self.operator == ComparisonOperator.GREATER_THAN:
            return environment_value > self.value

        raise UnreachableCodeError


@dataclass(frozen=True)  # TODO (py3.9): Use slots=True
class Marker:
    boolean: BooleanOperator | None
    markers: tuple[Marker | MarkerAtom, ...]

    @classmethod
    def from_string(cls, marker_string: str) -> Marker:
        return MarkerParser(marker_string).parse()

    def __str__(self) -> str:
        if not self.markers:
            return ""

        if self.boolean is None:
            return str(self.markers[0])

        joiner = f" {self.boolean} "
        return joiner.join(
            str(marker) if isinstance(marker, MarkerAtom) else f"({marker})"
            for marker in self.markers
        )

    @classmethod
    def from_dict(cls, specs: dict[str, Any]) -> Marker:  # TODO (py3.10): Use Self
        boolean = specs.get("boolean")
        if isinstance(boolean, str):
            boolean = BooleanOperator(boolean)
        markers: list[Marker | MarkerAtom] = []
        if "markers" not in specs:
            return cls(boolean=None, markers=(MarkerAtom.from_dict(specs),))
        for marker in specs["markers"]:
            variable = marker.get("variable")
            if isinstance(variable, str):
                variable = MarkerVariable(variable)
            if isinstance(variable, MarkerVariable):  # TODO (py3.9): Use match
                markers.append(MarkerAtom.from_dict(marker))
            else:
                markers.append(cls.from_dict(marker))

        return cls(boolean=boolean, markers=tuple(markers))

    def evaluate(self, extra: str = "") -> bool:
        if not self.markers:
            return True

        if self.boolean == BooleanOperator.OR:  # TODO (py3.9): Use match
            return any(marker.evaluate(extra) for marker in self.markers)
        if self.boolean == BooleanOperator.AND:
            return all(marker.evaluate(extra) for marker in self.markers)
        return self.markers[0].evaluate(extra)


@dataclass(frozen=True)  # TODO (py3.9): Use slots=True
class Token:
    name: re.Pattern[str]
    text: str
    position: int


class MarkerParser:
    def __init__(self, marker_string: str) -> None:
        self.string = marker_string
        self.next_token: Token | None = None
        self.position = 0

    def parse(self) -> Marker:
        if not self.string:
            return Marker(boolean=None, markers=())

        marker = self.to_dict(self.parse_marker())
        if isinstance(marker, MarkerAtom):
            return Marker(boolean=None, markers=(marker,))
        return marker

    @classmethod
    def to_dict(cls, parsed_marker: list[Any] | MarkerAtom) -> Marker | MarkerAtom:
        if isinstance(parsed_marker, MarkerAtom):
            return parsed_marker

        if len(parsed_marker) == 1:
            return cls.to_dict(parsed_marker[0])

        operators = {
            operator for index, operator in enumerate(parsed_marker) if index % 2
        }
        if len(operators) == 1:
            return Marker(
                boolean=parsed_marker[1],
                markers=tuple(
                    cls.to_dict(sub_marker)
                    for index, sub_marker in enumerate(parsed_marker)
                    if not index % 2
                ),
            )

        grouped_markers = cls.group_boolean_operators(parsed_marker)
        return cls.to_dict(grouped_markers)

    @staticmethod
    def group_boolean_operators(ungrouped: list[Any]) -> list[Any]:
        output: list[Any] = []
        tmp: list[Any] = []
        for i in range(len(ungrouped) // 2):
            op = ungrouped[2 * i + 1]
            if op == BooleanOperator.OR and tmp:
                tmp.append(ungrouped[2 * i])
                output.extend([tmp, BooleanOperator.OR])
                tmp = []
            elif op == BooleanOperator.OR:
                output.extend([ungrouped[2 * i], BooleanOperator.OR])
            else:
                tmp.extend([ungrouped[2 * i], BooleanOperator.AND])
        if tmp:
            tmp.append(ungrouped[-1])
            output.append(tmp)
        else:
            output.append(ungrouped[-1])
        return output

    def parse_marker(self) -> list[Any]:
        expression: list[Any] = [self.parse_marker_atom()]
        while self.check(TokenRule.BOOLEAN_OPERATOR):
            token = self.read()
            expr_right = self.parse_marker_atom()
            expression.extend([BooleanOperator(token.text), expr_right])
        return expression

    def parse_marker_atom(self) -> list[Any] | MarkerAtom:
        marker: list[Any] | MarkerAtom
        self.consume(TokenRule.WHITESPACE)
        if self.check(TokenRule.LEFT_PARENTHESIS, prepare_token=False):
            with self.enclosing_tokens(
                TokenRule.LEFT_PARENTHESIS, TokenRule.RIGHT_PARENTHESIS
            ):
                self.consume(TokenRule.WHITESPACE)
                marker = self.parse_marker()
                self.consume(TokenRule.WHITESPACE)
        else:
            marker = self.parse_marker_item()
        self.consume(TokenRule.WHITESPACE)
        return marker

    def parse_marker_item(self) -> MarkerAtom:
        self.consume(TokenRule.WHITESPACE)
        marker_left = self.parse_marker_var_or_str()
        self.consume(TokenRule.WHITESPACE)
        marker_op = self.parse_marker_op()
        self.consume(TokenRule.WHITESPACE)
        marker_right = self.parse_marker_var_or_str()
        self.consume(TokenRule.WHITESPACE)

        if isinstance(marker_right, MarkerVariable):
            marker_left, marker_right = marker_right, marker_left
        if not (
            isinstance(marker_left, MarkerVariable) and isinstance(marker_right, str)
        ):
            msg = "Expected a marker variable and a string"
            raise TypeError(msg)

        if marker_left == MarkerVariable.EXTRA:
            marker_right = canonicalise_name(marker_right)
        return MarkerAtom(variable=marker_left, operator=marker_op, value=marker_right)

    def parse_marker_op(self) -> ComparisonOperator:
        if self.check(TokenRule.IN):
            self.read()
            return ComparisonOperator.IN
        if self.check(TokenRule.NOT_IN):
            self.read()
            return ComparisonOperator.NOT_IN
        if self.check(TokenRule.OPERATOR):
            return ComparisonOperator(self.read().text)
        msg = "Expected a marker operator."
        raise ValueError(msg)

    def parse_marker_var_or_str(self) -> MarkerVariable | str:
        if self.check(TokenRule.VARIABLE):
            return MarkerVariable(self.read().text.replace(".", "_"))
        if self.check(TokenRule.QUOTED_STRING):
            return str(ast.literal_eval(self.read().text))
        msg = "Expected a marker variable"
        raise ValueError(msg)

    def consume(self, name: TokenRule) -> None:
        if self.check(name):
            self.read()

    def check(self, rule: TokenRule, *, prepare_token: bool = True) -> bool:
        expression = rule.value

        match = expression.match(self.string, self.position)
        if match is None:
            return False

        if prepare_token:
            self.next_token = Token(expression, match[0], self.position)

        return True

    def read(self) -> Token:
        token = self.next_token

        if token is None:
            msg = "No token to read"
            raise RuntimeError(msg)

        self.position += len(token.text)
        self.next_token = None

        return token

    @contextmanager
    def enclosing_tokens(
        self, open_token: TokenRule, close_token: TokenRule
    ) -> Iterator[None]:
        if self.check(open_token):
            open_position = self.position
            self.read()
        else:
            open_position = None

        yield

        if open_position is None:
            return

        if not self.check(close_token):
            msg = f"Expected matching {close_token} for {open_token}"
            raise RuntimeError(msg)

        self.read()



from __future__ import annotations

import csv
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any
from zipfile import ZIP_DEFLATED, ZipFile

from phosphorus.__version__ import __version__
from phosphorus.construction.base import Builder
from phosphorus.lib.licenses import get_license_files
from phosphorus.lib.tags import Tag
from phosphorus.lib.zipped_file import ArchiveFile

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator


class WheelBuilder(Builder):
    __slots__ = ("editable",)

    def __init__(
        self,
        output_dir: Path,
        config: dict[str, Any] | None,
        metadata_dir: Path | None,
        *,
        editable: bool = False,
    ) -> None:
        super().__init__(output_dir, config, metadata_dir)
        self.editable = editable

    @property
    def filename(self) -> str:
        tag = Tag(
            interpreter=self.config.get("interpreter", "py3"),
            abi=self.config.get("abi"),
            platform=self.config.get("platform", "any"),
        )
        return self.wheel_filenames[tag]

    @property
    def dist_info(self) -> str:
        return f"{self.base_name}.dist-info"

    @property
    def wheel_filenames(self) -> dict[Tag, str]:
        return {tag: f"{self.base_name}-{tag}.whl" for tag in self.meta.tags}

    def package_files(self, temp_dir: Path) -> Iterator[ArchiveFile]:
        if self.editable:
            file = self.create_pth(temp_dir)
            yield ArchiveFile.from_file(source=file, base_dir=temp_dir)
            return

        for package in self.meta.package_paths:
            for file in package.absolute_path.rglob("*"):
                if file.is_file():
                    yield ArchiveFile.from_file(
                        source=file, base_dir=package.absolute_path
                    )

    def non_package_files(self, temp_dir: Path) -> Iterator[ArchiveFile]:
        for file in self.prepare_metadata(temp_dir).rglob("*"):
            if file.is_file():
                yield ArchiveFile.from_file(source=file, base_dir=temp_dir)

    def get_info_file(self, temp_dir: Path, data: list[list[Any]]) -> ArchiveFile:
        info_file = temp_dir.joinpath(self.dist_info, "RECORD")
        with info_file.open("w") as csv_file:
            write = csv.writer(csv_file)
            write.writerows(data)
            write.writerow([info_file.relative_to(temp_dir), "", ""])

        return ArchiveFile.from_file(source=info_file, base_dir=temp_dir)

    def write_files(
        self, files: Iterable[ArchiveFile], package: Path, temp_dir: Path
    ) -> None:
        with ZipFile(package, mode="w", compression=ZIP_DEFLATED) as zip_file:
            rows = []
            for archive_file in sorted(files):
                zip_file.writestr(
                    archive_file.zip_info,
                    archive_file.absolute_path.read_bytes(),
                    compress_type=ZIP_DEFLATED,
                )
                rows.append(
                    [archive_file.relative_path, archive_file.digest, archive_file.size]
                )

            record_info = self.get_info_file(temp_dir, rows)
            zip_file.writestr(
                record_info.zip_info,
                record_info.absolute_path.read_bytes(),
                compress_type=ZIP_DEFLATED,
            )

    @property
    def record_target(self) -> Path:
        return Path(self.dist_info).joinpath("RECORD")

    def is_pure_lib(self) -> bool:
        return all(tag.abi is None for tag in self.meta.tags)

    def create_pth(self, tmp_dir: Path) -> Path:
        paths = {
            package.absolute_path.as_posix() for package in self.meta.package_paths
        }
        pth = tmp_dir.joinpath(f"{self.meta.package.name}.pth")
        pth.write_text("\n".join(sorted(paths)))
        return pth

    def prepare_metadata(self, temp_dir: Path | None = None) -> Path:
        base = temp_dir or self.metadata_dir
        if base is None:
            msg = f"Cannot create {self.dist_info} without a metadata directory."
            raise TypeError(msg)

        dist_info = base.joinpath(self.dist_info)
        shutil.rmtree(dist_info, ignore_errors=True)
        dist_info.mkdir(parents=True, exist_ok=True)

        for path, content_generator in self.get_dist_info_entries():
            with dist_info.joinpath(path).open("w") as file:
                file.writelines(f"{line}\n" for line in content_generator)

        for license_file in get_license_files(self.meta.base_dir):
            destination = dist_info.joinpath(
                license_file.relative_to(self.meta.base_dir)
            )

            destination.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(license_file, destination)

        return dist_info

    def get_dist_info_entries(self) -> Iterator[tuple[str, Iterator[str]]]:
        yield "WHEEL", self.get_wheel_content()
        yield "METADATA", self.get_metadata_content()
        if self.meta.scripts:
            yield "entry_points.txt", self.get_entry_points_content()

    def get_entry_points_content(self) -> Iterator[str]:
        yield "[console_scripts]"
        for script in self.meta.scripts:
            yield f"{script.command}={script.entrypoint}"

    def get_wheel_content(self) -> Iterator[str]:
        yield "Wheel-Version: 1.0"
        yield f"Generator: phosphorus {__version__}"
        pure_lib = str(self.is_pure_lib()).lower()
        yield f"Root-Is-Purelib: {pure_lib}"
        for tag in self.meta.tags:
            yield f"Tag: {tag}"



from __future__ import annotations

from typing import TYPE_CHECKING, TextIO

from phosphorus.__version__ import __version__
from phosphorus._seven import TOMLDecodeError
from phosphorus.lib.resolver import LockEntry, Resolver
from phosphorus.lib.term import SGRParams, SGRString
from phosphorus.subcommands.base import BaseCommand

if TYPE_CHECKING:
    from argparse import Namespace


class LockCommand(BaseCommand):
    __slots__ = (
        "force",
        "enforce_pep440",
        "allow_pre_releases",
        "allow_dev_releases",
        "timestamp",
    )

    def __init__(self, args: Namespace, /) -> None:
        super().__init__(args)
        self.force = args.force
        self.enforce_pep440 = args.enforce_pep440
        self.allow_pre_releases = args.allow_pre_releases
        self.allow_dev_releases = args.allow_dev_releases

    def run(self) -> None:
        if self.force:
            print(
                f"🔄🔒 {SGRString('lock', params=[SGRParams.GREEN])} "
                "was run with --force, rebuilding the lock ..."
            )
            dependencies = self.resolve_dependencies()
            return self.write_lockfile(dependencies)

        try:
            p_hash = self.get_current_hash()
        except (FileNotFoundError, KeyError, TOMLDecodeError):
            p_hash = None

        if p_hash == self.meta.p_hash:
            print("🔒✅ Lockfile is up to date, nothing to do ...")
            return None

        print("🔄 Lockfile is not up to date, rebuilding the lock ...")
        dependencies = self.resolve_dependencies()
        return self.write_lockfile(dependencies)

    def resolve_dependencies(self) -> list[LockEntry]:
        print("🔍 Resolving dependencies ...")
        return Resolver(
            self.meta.requirement_groups,
            enforce_pep440=self.enforce_pep440,
            allow_pre_releases=self.allow_pre_releases,
            allow_dev_releases=self.allow_dev_releases,
            verbosity=self.verbosity,
        ).resolve()

    def write_lockfile(self, dependencies: list[LockEntry]) -> None:
        print("🔒 Writing new lockfile ...")
        with self.meta.lockfile.open("w") as lockfile:
            lockfile.write('["$meta"]\n')
            lockfile.write(f'version = "{__version__}"\n')
            lockfile.write(f'hash = "{self.meta.p_hash}"\n')
            for dependency in sorted(dependencies):
                self.write_dependency_to_lockfile(dependency, lockfile)

    @staticmethod
    def write_dependency_to_lockfile(entry: LockEntry, lockfile: TextIO) -> None:
        requirement = entry.requirement
        lockfile.write("\n[[packages]]\n")
        lockfile.write(f'name = "{requirement.package}"\n')
        lockfile.write(f'version = "{requirement.version}"\n')
        lockfile.write("groups = [\n")
        for group in entry.groups:
            lockfile.write(f'    "{group}",\n')
        lockfile.write("]\n")
        if marker := str(requirement.marker):
            lockfile.write(f'marker = "{marker}"\n')
        lockfile.write("hashes = [\n")
        for file_hash in entry.hashes:
            lockfile.write(f'    "{file_hash}",\n')
        lockfile.write("]\n")



from __future__ import annotations

from typing import TYPE_CHECKING

from phosphorus.lib.constants import pyproject_base_name

if TYPE_CHECKING:
    from pathlib import Path


class UnreachableCodeError(AssertionError):
    """This is an aid for static analysers"""


class MissingProjectRootError(RuntimeError):
    """The project root could not be found"""

    def __init__(self, path: Path) -> None:
        msg = f"Could not find {pyproject_base_name} in {path} or any parent directory"
        super().__init__(msg)


class ImproperlyConfiguredProjectError(RuntimeError):
    """The pyproject.toml file is not correctly formatted"""

    def __init__(self, key: str) -> None:
        msg = f"The pyproject.toml file is missing the key: {key} and isn't marked as dynamic"
        super().__init__(msg)


class ThirdPartyError(RuntimeError):
    """A third party cli tool exited unsuccessfully"""

    def __init__(self, command: str) -> None:
        msg = f"Command `{command}` exited with a non-zero status code"
        super().__init__(msg)



from __future__ import annotations

import os
import platform
import re
import sys
from enum import Enum, unique
from pathlib import Path

package_cache = Path.home().joinpath(".cache/phosphorus/packages/")
pypi_cache = package_cache.joinpath("PyPI")
pyproject_base_name = "pyproject.toml"
lock_file_name = "p-lock.toml"
hash_prefix = "--hash=sha256:"

# Accept the British spelling for the noun
licence_base_names = {"COPYING", "LICENCE", "LICENCES", "LICENSE", "LICENSES"}


@unique
class ComparisonOperator(Enum):
    EXACT_MATCH = "==="
    EQUAL_TO = "=="
    NOT_EQUAL = "!="
    COMPATIBLE_WITH = "~="
    LESS_OR_EQUAL = "<="
    GREATER_OR_EQUAL = ">="
    LESS_THAN = "<"
    GREATER_THAN = ">"
    IN = "in"
    NOT_IN = "not in"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def prefix_match_operator(cls) -> set[ComparisonOperator]:
        return {cls.EQUAL_TO, cls.NOT_EQUAL}

    @classmethod
    def env_marker_operator(cls) -> set[ComparisonOperator]:
        return {cls.IN, cls.NOT_IN}

    @classmethod
    def non_local_operator(cls) -> set[ComparisonOperator]:
        return {
            cls.LESS_THAN,
            cls.LESS_OR_EQUAL,
            cls.GREATER_OR_EQUAL,
            cls.GREATER_THAN,
            cls.COMPATIBLE_WITH,
        }


@unique
class BooleanOperator(Enum):
    AND = "and"
    OR = "or"

    def __str__(self) -> str:
        return self.value


@unique
class MarkerVariable(Enum):
    PYTHON_VERSION = "python_version"
    PYTHON_FULL_VERSION = "python_full_version"
    OS_NAME = "os_name"
    SYS_PLATFORM = "sys_platform"
    PLATFORM_RELEASE = "platform_release"
    PLATFORM_SYSTEM = "platform_system"
    PLATFORM_VERSION = "platform_version"
    PLATFORM_MACHINE = "platform_machine"
    PLATFORM_PYTHON_IMPLEMENTATION = "platform_python_implementation"
    IMPLEMENTATION_NAME = "implementation_name"
    IMPLEMENTATION_VERSION = "implementation_version"
    EXTRA = "extra"

    def __str__(self) -> str:
        return self.value

    def get_env_value(self, extra: str = "") -> str:
        if self == MarkerVariable.PYTHON_VERSION:  # TODO (py3.9): Use match
            return ".".join(platform.python_version_tuple()[:2])
        if self == MarkerVariable.PYTHON_FULL_VERSION:
            return platform.python_version()
        if self == MarkerVariable.OS_NAME:
            return os.name
        if self == MarkerVariable.SYS_PLATFORM:
            return sys.platform
        if self == MarkerVariable.PLATFORM_RELEASE:
            return platform.release()
        if self == MarkerVariable.PLATFORM_SYSTEM:
            return platform.system()
        if self == MarkerVariable.PLATFORM_VERSION:
            return platform.version()
        if self == MarkerVariable.PLATFORM_MACHINE:
            return platform.machine()
        if self == MarkerVariable.PLATFORM_PYTHON_IMPLEMENTATION:
            return platform.python_implementation()
        if self == MarkerVariable.IMPLEMENTATION_NAME:
            if not hasattr(sys, "implementation"):
                return ""
            return sys.implementation.name
        if self == MarkerVariable.IMPLEMENTATION_VERSION:
            if not hasattr(sys, "implementation"):
                return "0"
            info = sys.implementation.version
            version = f"{info.major}.{info.minor}.{info.micro}"
            if (kind := info.releaselevel) == "final":
                return version
            return f"{version}{kind[0]}{info.serial}"
        return extra


class TokenRule(Enum):
    LEFT_PARENTHESIS = re.compile(r"\(")
    RIGHT_PARENTHESIS = re.compile(r"\)")
    WHITESPACE = re.compile(r"[ \t]+")
    OPERATOR = re.compile(r"(===|==|~=|!=|<=|>=|<|>)")
    BOOLEAN_OPERATOR = re.compile(r"\b(or|and)\b")
    IN = re.compile(r"\bin\b")
    NOT_IN = re.compile(r"\bnot[ \t]+in\b")
    QUOTED_STRING = re.compile(
        r"""
            (
                ('[^']*')
                |
                ("[^"]*")
            )
        """,
        re.VERBOSE,
    )
    VARIABLE = re.compile(
        r"""
            \b(
                python_version
                |python_full_version
                |os[._]name
                |sys[._]platform
                |platform_(release|system)
                |platform[._](version|machine|python_implementation)
                |python_implementation
                |implementation_(name|version)
                |extra
            )\b
        """,
        re.VERBOSE,
    )



from __future__ import annotations

from dataclasses import dataclass, field
from functools import total_ordering
from itertools import dropwhile
from typing import TYPE_CHECKING, Any

from phosphorus.lib.constants import ComparisonOperator
from phosphorus.lib.exceptions import UnreachableCodeError
from phosphorus.lib.regex import version_pattern, version_separators

if TYPE_CHECKING:
    from phosphorus.lib.types import Match


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Epoch:
    epoch: int

    @classmethod
    def from_string(cls, match: Match) -> Epoch:  # TODO (py3.10): Use Self
        return cls(epoch=int(match or 0))

    def __str__(self) -> str:
        return f"{self.epoch}!" if self.epoch else ""


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Release:
    release: tuple[int, ...]
    full_release: tuple[int, ...] = field(repr=False, compare=False)

    def __init__(self, full_release: tuple[int, ...]) -> None:
        release = tuple(
            reversed(list(dropwhile(lambda x: x == 0, reversed(full_release))))
        )
        object.__setattr__(self, "release", release)
        object.__setattr__(self, "full_release", full_release)

    @classmethod
    def from_string(cls, match: str) -> Release:  # TODO (py3.10): Use Self
        full_release = tuple(int(segment) for segment in match.split("."))
        return cls(full_release=full_release)

    def __str__(self) -> str:
        return ".".join(str(x) for x in self.full_release)

    @property
    def canonical_form(self) -> str:
        return ".".join(str(x) for x in self.release)

    @property
    def major(self) -> int:
        return self.release[0]

    @property
    def minor(self) -> int:
        try:
            return self.release[1]
        except IndexError:
            return 0

    @property
    def micro(self) -> int:
        try:
            return self.release[2]
        except IndexError:
            return 0

    def padded(self, zeroes: int) -> Release:  # TODO (py3.10): Use Self
        n = len(self.release)
        if zeroes < n:
            msg = "Padding cannot truncate the release"
            raise ValueError(msg)
        if zeroes == n:
            return self
        full_release = self.release + (0,) * (zeroes - n)
        return self.__class__(full_release)


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Pre:
    letter: str
    number: int

    @classmethod
    def from_string(
        cls, match_letter: Match, match_number: Match
    ) -> Pre:  # TODO (py3.10): Use Self
        number = int(match_number or 0)
        letter = cls._canonicalize_letter(match_letter) if match_letter else "z"
        return cls(letter=letter, number=number)

    def __bool__(self) -> bool:
        return self.letter != "z"

    def __str__(self) -> str:
        return f"{self.letter}{self.number}" if self else ""

    @staticmethod
    def _canonicalize_letter(letter: str) -> str:
        letter = letter.lower()

        if letter == "alpha":
            return "a"
        if letter == "beta":
            return "b"
        if letter in {"c", "pre", "preview"}:
            return "rc"

        return letter


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Post:
    post: int

    @classmethod
    def from_string(
        cls, match_letter: Match, match_number: Match
    ) -> Post:  # TODO (py3.10): Use Self
        return cls(post=int(match_number or 0) if match_letter or match_number else -1)

    def __bool__(self) -> bool:
        return self.post >= 0

    def __str__(self) -> str:
        return f".post{self.post}" if self else ""


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Dev:
    dev: int | float

    @classmethod
    def from_string(
        cls, match_letter: Match, match_number: Match
    ) -> Dev:  # TODO (py3.10): Use Self
        return cls(dev=int(match_number or 0) if match_letter else float("inf"))

    def __bool__(self) -> bool:
        return self.dev != float("inf")

    def __str__(self) -> str:
        return f".dev{self.dev}" if self else ""


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class Local:
    local: tuple[str | int, ...]

    @classmethod
    def from_string(cls, match: Match) -> Local:  # TODO (py3.10): Use Self
        local_version = (
            tuple(
                int(part) if part.isdigit() else part.lower()
                for part in version_separators.split(match)
            )
            if match
            else ()
        )
        return cls(local=local_version)

    def __bool__(self) -> bool:
        return bool(self.local)

    def __str__(self) -> str:
        if not self.local:
            return ""
        canonical_local = ".".join(str(part) for part in self.local)
        return f"+{canonical_local}"


@total_ordering
@dataclass(frozen=True)  # TODO (py3.9): Use slots=True
class Version:
    epoch: Epoch
    release: Release
    post: Post
    pre: Pre
    dev: Dev
    local: Local
    prefix_match: bool = field(repr=False, default=False)
    match_all: bool = field(repr=False, default=False)
    pep_440_compliant: bool = field(repr=False, default=True)

    @classmethod
    def from_string(cls, version: str) -> Version:  # TODO (py3.10): Use Self
        if version == "*" or version.endswith("!*"):
            epoch = int(version[:-2]) if version.endswith("!*") else -1
            return cls(
                epoch=Epoch.from_string(str(epoch)),
                release=Release.from_string("0"),
                pre=Pre.from_string(None, None),
                post=Post.from_string(None, None),
                dev=Dev.from_string(None, None),
                local=Local.from_string(None),
                prefix_match=True,
                match_all=True,
            )
        if version.endswith(".*"):
            prefix_match = True
            version = version[:-2]
        else:
            prefix_match = False
        match = version_pattern.search(version.strip())
        if not match:
            return cls(
                epoch=Epoch(epoch=-2),
                release=Release.from_string("0"),
                pre=Pre.from_string(None, None),
                post=Post.from_string(None, None),
                dev=Dev.from_string(None, None),
                local=Local.from_string(version),
                pep_440_compliant=False,
            )

        if prefix_match and (match.group("dev_l") or match.group("local")):
            msg = "Prefix match containing a dev or local release is invalid"
            raise RuntimeError(msg)
        return cls(
            epoch=Epoch.from_string(match.group("epoch")),
            release=Release.from_string(match.group("release")),
            pre=Pre.from_string(match.group("pre_l"), match.group("pre_n")),
            post=Post.from_string(
                match.group("post_l"), match.group("post_n1") or match.group("post_n2")
            ),
            dev=Dev.from_string(match.group("dev_l"), match.group("dev_n")),
            local=Local.from_string(match.group("local")),
            prefix_match=prefix_match,
        )

    def __repr__(self) -> str:
        pre = (self.pre.letter, self.pre.number) if self.is_pre_release else None
        post = self.post.post if self.is_post_release else None
        dev = self.dev.dev if self.is_dev_release else None
        return (
            "Version("
            f"epoch={self.epoch.epoch}, "
            f"release={self.release.full_release}, "
            f"pre={pre}, "
            f"post={post}, "
            f"dev={dev}, "
            f"local={self.local.local or None}"
            ")"
        )

    def __str__(self) -> str:
        return f"{self.epoch}{self.release}{self.pre}{self.post}{self.dev}{self.local}"

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        if self.epoch != other.epoch:
            return self.epoch < other.epoch

        if self.release != other.release:
            return self.release < other.release

        if self.post != other.post:
            return self.post < other.post

        if self.is_dev_release ^ other.is_dev_release:
            return self.is_dev_release

        if self.pre != other.pre:
            return self.pre < other.pre

        if self.dev != other.dev:
            return self.dev < other.dev

        return self.local < other.local

    @property
    def major(self) -> int:
        return self.release.major

    @property
    def minor(self) -> int:
        return self.release.minor

    @property
    def micro(self) -> int:
        return self.release.micro

    @property
    def is_post_release(self) -> bool:
        return bool(self.post)

    @property
    def is_pre_release(self) -> bool:
        return bool(self.pre)

    @property
    def is_dev_release(self) -> bool:
        return bool(self.dev)

    @property
    def is_base_version(self) -> bool:
        return not (self.post or self.pre or self.dev or self.local)

    @property
    def base_version(self) -> Version:  # TODO (py3.10): Use Self
        if self.is_base_version:
            return self
        return self.__class__.from_string(f"{self.epoch}{self.release}")

    @property
    def canonical_form(self) -> str:
        return f"{self.epoch}{self.release.canonical_form}{self.pre}{self.post}{self.dev}{self.local}"

    @property
    def public_version(self) -> Version:  # TODO (py3.10): Use Self
        return self.__class__.from_string(
            f"{self.epoch}{self.release.canonical_form}{self.pre}{self.post}{self.dev}"
        )

    def padded(self, zeroes: int) -> Version:  # TODO (py3.10): Use Self
        release = self.release.padded(zeroes)
        if release.full_release == self.release.full_release:
            return self
        return self.__class__(
            epoch=self.epoch,
            release=release,
            post=self.post,
            pre=self.pre,
            dev=self.dev,
            local=self.local,
            prefix_match=self.prefix_match,
            match_all=self.match_all,
        )


@dataclass(frozen=True, order=True)  # TODO (py3.9): Use slots=True
class VersionClause:
    operator: ComparisonOperator
    identifier: Version

    def __post_init__(self) -> None:
        if self.operator in ComparisonOperator.env_marker_operator():
            msg = f"Only environment markers are permitted to use `{self.operator}`"
            raise RuntimeError(msg)
        if not self.identifier.pep_440_compliant:
            if self.operator != ComparisonOperator.EXACT_MATCH:
                msg = "Non PEP-440 versions can only be compared with ==="
                raise RuntimeError(msg)
            return
        if self.operator in ComparisonOperator.non_local_operator():
            if self.identifier.local:
                msg = "Local version identifiers are not permitted with this type of clause"
                raise RuntimeError(msg)
            if (
                self.operator == ComparisonOperator.COMPATIBLE_WITH
                and len(self.identifier.release.full_release) == 1
            ):
                msg = (
                    "Compatible release operator is not permitted with single segments"
                )
                raise RuntimeError(msg)
        if (
            self.operator not in ComparisonOperator.prefix_match_operator()
            and self.identifier.prefix_match
        ):
            msg = "Prefix match is only allowed in match or exclusion"
            raise RuntimeError(msg)

    def __str__(self) -> str:
        return f"{self.operator}{self.identifier}"

    @classmethod
    def from_string(cls, clause: str) -> VersionClause:  # TODO (py3.10): Use Self
        for candidate in ComparisonOperator:
            if clause.startswith(candidate.value):
                operator_length = len(candidate.value)
                operator = candidate
                break
        else:
            operator_length = 0
            operator = ComparisonOperator.EQUAL_TO

        return cls(
            operator=operator, identifier=Version.from_string(clause[operator_length:])
        )

    def match(self, candidate: Version) -> bool:
        if not (self.identifier.pep_440_compliant and candidate.pep_440_compliant):
            return self.match_exact(candidate)

        if (
            self.operator == ComparisonOperator.COMPATIBLE_WITH
        ):  # TODO (py3.9): Use match
            return self.match_compatible(candidate)
        if self.operator == ComparisonOperator.EQUAL_TO:
            return self.match_equality(candidate)
        if self.operator == ComparisonOperator.NOT_EQUAL:
            return not self.match_equality(candidate)
        if self.operator == ComparisonOperator.LESS_OR_EQUAL:
            return self.match_leq(candidate)
        if self.operator == ComparisonOperator.GREATER_OR_EQUAL:
            return self.match_geq(candidate)
        if self.operator == ComparisonOperator.LESS_THAN:
            return self.match_lt(candidate)
        if self.operator == ComparisonOperator.GREATER_THAN:
            return self.match_gt(candidate)
        if self.operator == ComparisonOperator.EXACT_MATCH:
            return self.match_exact(candidate)

        raise UnreachableCodeError

    def match_compatible(self, candidate: Version) -> bool:
        if not self.match_geq(candidate):
            return False

        release_parts = str(self.identifier.release).split(".")
        release_parts[-1] = "*"
        wildcard = Version.from_string(
            f"{self.identifier.epoch}{'.'.join(release_parts)}"
        )
        return VersionClause(
            operator=ComparisonOperator.EQUAL_TO, identifier=wildcard
        ).match(candidate)

    def match_equality(self, candidate: Version) -> bool:
        if candidate.epoch != self.identifier.epoch:
            return self.identifier.epoch.epoch == -1

        if self.identifier.match_all:
            return True

        if not self.identifier.prefix_match:
            if self.identifier.local.local:
                return candidate == self.identifier
            return candidate.public_version == self.identifier.public_version

        if not self.identifier.is_base_version:
            return candidate.canonical_form.startswith(self.identifier.canonical_form)

        if not candidate.is_base_version:
            candidate = candidate.base_version

        n = len(self.identifier.release.full_release)
        if len(candidate.release.full_release) <= n:
            candidate = candidate.padded(n)
        return str(candidate).startswith(str(self.identifier))

    def match_leq(self, candidate: Version) -> bool:
        return candidate <= self.identifier

    def match_geq(self, candidate: Version) -> bool:
        return candidate >= self.identifier

    def match_lt(self, candidate: Version) -> bool:
        candidate = candidate.public_version

        if self.identifier.is_pre_release:
            return candidate < self.identifier

        return candidate.base_version < self.identifier

    def match_gt(self, candidate: Version) -> bool:
        candidate = candidate.public_version

        if self.identifier.is_post_release:
            return candidate > self.identifier

        return candidate.base_version > self.identifier

    def match_exact(self, candidate: Version) -> bool:
        return candidate == self.identifier



from __future__ import annotations

from subprocess import run
from typing import TYPE_CHECKING

from uv import find_uv_bin

if TYPE_CHECKING:
    from subprocess import CompletedProcess


def uv_run(args: list[str]) -> CompletedProcess[bytes]:
    uv_bin = find_uv_bin()
    return run([uv_bin, *args], capture_output=True)  # noqa: PLW1510, S603



from __future__ import annotations

import tarfile
from typing import TYPE_CHECKING

from phosphorus.construction.base import Builder
from phosphorus.lib.constants import pyproject_base_name
from phosphorus.lib.licenses import get_license_files
from phosphorus.lib.zipped_file import ArchiveFile

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from pathlib import Path


class SdistBuilder(Builder):
    __slots__ = ()

    @property
    def filename(self) -> str:
        return f"{self.base_name}.tar.gz"

    def package_files(self, _temp_dir: Path) -> Iterator[ArchiveFile]:
        base_dir = self.meta.base_dir
        for package in self.meta.package_paths:
            for file in package.absolute_path.rglob("*"):
                if file.is_file():
                    yield ArchiveFile.from_file(source=file, base_dir=base_dir)

    def non_package_files(self, _temp_dir: Path) -> Iterator[ArchiveFile]:
        base_dir = self.meta.base_dir
        yield ArchiveFile.from_file(
            base_dir.joinpath(pyproject_base_name), base_dir=base_dir
        )

        for license_file in get_license_files(base_dir):
            yield ArchiveFile.from_file(license_file, base_dir=base_dir)

        if self.meta.readme.read_text():
            yield ArchiveFile.from_file(self.meta.readme, base_dir=base_dir)

    def get_info_file(self, temp_dir: Path, _data: str = "") -> ArchiveFile:
        pkg_info = temp_dir.joinpath("PKG-INFO")
        with pkg_info.open("w") as file:
            file.writelines(f"{line}\n" for line in self.get_metadata_content())

        return ArchiveFile.from_file(source=pkg_info, base_dir=temp_dir)

    def write_files(
        self, files: Iterable[ArchiveFile], package: Path, temp_dir: Path
    ) -> None:
        with tarfile.open(package, "w:gz") as tar:
            for archive_file in sorted(files):
                tar.addfile(
                    archive_file.tar_info, archive_file.absolute_path.open("rb")
                )

            info_file = self.get_info_file(temp_dir)
            tar.addfile(info_file.tar_info, info_file.absolute_path.open("rb"))



from __future__ import annotations

import sys
from argparse import Namespace
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING
from warnings import warn

from phosphorus._seven import toml_parser
from phosphorus.lib.constants import hash_prefix
from phosphorus.lib.exceptions import ThirdPartyError
from phosphorus.lib.markers import Marker
from phosphorus.lib.packages import Package
from phosphorus.lib.requirements import ResolvedRequirement
from phosphorus.lib.resolver import LockEntry
from phosphorus.lib.term import SGRParams, SGRString
from phosphorus.lib.uv_runner import uv_run
from phosphorus.lib.versions import Version
from phosphorus.subcommands.base import BaseCommand
from phosphorus.subcommands.lock import LockCommand

if TYPE_CHECKING:
    from phosphorus.lib.types import PackageDiff


class InstallCommand(BaseCommand):
    __slots__ = ("groups", "sync")

    def __init__(self, args: Namespace, /) -> None:
        super().__init__(args)
        self.sync = args.sync
        self.groups = self.get_groups(set(args.groups), set(args.exclude))

    def run(self) -> None:
        try:
            self.install()
        except RuntimeError as exc:
            print(f"❌ Installation failed: {exc}", file=sys.stderr)
            sys.exit(1)
        else:
            print("✅ Success: Installation finished successfully!")

    def install(self) -> None:
        if not self.groups:
            msg = "No remaining valid groups"
            raise RuntimeError(msg)

        if not self.meta.lockfile.exists():
            print("🔒 Lockfile not found, creating one...")
            LockCommand(
                Namespace(
                    enforce_pep440=True,
                    allow_pre_releases=False,
                    allow_dev_releases=False,
                    force=False,
                )
            ).run()

        package_diff = self.get_package_diff()

        if package_diff["update"]:
            print("📦 Installing packages...")
        else:
            print("📦 No packages to be updated, skipping...")
        for old_package, lock_entry in package_diff["update"]:
            self.install_package(old_package, lock_entry)
        if self.sync:
            if package_diff["remove"]:
                print("🧹 Removing packages...")
            else:
                print("🧹 No packages to be removed, skipping...")
            for old_package in package_diff["remove"]:
                self.remove_package(old_package)
        self.install_self()

    def get_groups(self, included: set[str], excluded: set[str]) -> set[str]:
        all_groups = {group.group for group in self.meta.requirement_groups}
        if included:
            if not included.issubset(all_groups):
                msg = f"Unknown group(s) {included - all_groups}, ignoring..."
                warn(msg, RuntimeWarning, stacklevel=3)
            return included & all_groups
        if not excluded.issubset(all_groups):
            msg = f"Unknown group(s) {excluded - all_groups}, ignoring..."
            warn(msg, RuntimeWarning, stacklevel=3)
        return all_groups - excluded

    def get_package_diff(self) -> PackageDiff:
        print("🔎 Looking for packages to install/remove/upgrade...")

        output = uv_run(["pip", "freeze", "--exclude-editable"])
        existing_venv = {
            resolved_requirement.package: resolved_requirement
            for requirement in output.stdout.decode().splitlines()
            if (resolved_requirement := ResolvedRequirement.from_string(requirement))
        }

        with self.meta.lockfile.open("rb") as lockfile:
            lock = toml_parser(lockfile)

        target_venv = {}
        for lock_entry in lock["packages"]:
            marker = Marker.from_string(lock_entry.get("marker", ""))
            if not marker.evaluate():
                continue
            if not self.groups & set(lock_entry["groups"]):
                continue
            package = Package(lock_entry["name"])
            entry = LockEntry(
                requirement=ResolvedRequirement(
                    package=package,
                    version=Version.from_string(lock_entry["version"]),
                    marker=marker,
                ),
                groups=tuple(lock_entry["groups"]),
                hashes=tuple(lock_entry["hashes"]),
            )
            target_venv[package] = entry

        update: set[tuple[ResolvedRequirement | None, LockEntry]] = set()
        remove = set()

        for package in (
            existing_venv.keys() | target_venv.keys()
        ):  # TODO (py3.9): Remove .keys()
            if package not in target_venv:
                remove.add(existing_venv[package])
            elif package not in existing_venv:
                update.add((None, target_venv[package]))
            else:
                existing = existing_venv[package]
                target = target_venv[package]
                if existing.version != target.requirement.version:
                    update.add((existing, target))

        return {"update": update, "remove": remove}

    def install_package(
        self, old_package: ResolvedRequirement | None, lock_entry: LockEntry
    ) -> None:
        package = lock_entry.requirement.package
        old_version = old_package.version if old_package else None
        new_version = lock_entry.requirement.version
        hashes = " ".join(
            f"{hash_prefix}{hashed_file}" for hashed_file in lock_entry.hashes
        )
        if not old_version:
            action = "Installing"
            version = str(new_version)
        elif old_version < new_version:
            action = "Upgrading"
            version = f"{old_version} → {new_version}"
        else:
            action = "Downgrading"
            version = f"{old_version} → {new_version}"
        print(
            f"⏳ {action} {SGRString(str(package), params=[SGRParams.CYAN])} "
            f"({SGRString(version, params=[SGRParams.BLUE_BRIGHT])})...",
            end=" ",
        )
        with NamedTemporaryFile("w", delete=False) as tmp:
            line = f"{package}=={new_version} {hashes}"
            tmp.write(f"{line}\n")
            tmp.seek(0)
            output = uv_run(
                ["pip", "install", "--no-deps", "--require-hashes", "-r", tmp.name]
            )
            if output.returncode == 0:
                print("🗸")
            else:
                print()
                command = "uv pip install"
                raise ThirdPartyError(command)

    def remove_package(self, old_package: ResolvedRequirement) -> None:
        package = old_package.package
        version = old_package.version
        print(
            f"⏳ Removing {SGRString(str(package), params=[SGRParams.CYAN])} "
            f"({SGRString(str(version), params=[SGRParams.BLUE_BRIGHT])})...",
            end=" ",
        )
        output = uv_run(["pip", "uninstall", package.name])
        if output.returncode == 0:
            print("🗸")
        else:
            print()
            command = "uv pip uninstall"
            raise ThirdPartyError(command)

    def install_self(self) -> None:
        print(
            f"⏳ Installing {SGRString(str(self.meta.package), params=[SGRParams.CYAN])} "
            f"({SGRString(str(self.meta.version), params=[SGRParams.BLUE_BRIGHT])})...",
            end=" ",
        )
        output = uv_run(["pip", "install", "--no-deps", "--editable", "."])
        if output.returncode == 0:
            print("🗸")
        else:
            print()
            command = "uv pip install"
            raise ThirdPartyError(command)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from phosphorus._seven import toml_parser
from phosphorus.lib.metadata import Metadata

if TYPE_CHECKING:
    from argparse import Namespace


class BaseCommand:
    __slots__ = ("meta", "verbosity")

    def __init__(self, _args: Namespace, /) -> None:
        self.verbosity = _args.verbosity
        self.meta = Metadata.from_path()

    def run(self) -> None:
        msg = f"{self.__class__.__qualname__} must implement run"
        raise NotImplementedError(msg)

    def get_current_hash(self) -> str:
        with self.meta.lockfile.open("rb") as file:
            return cast(str, toml_parser(file)["$meta"]["hash"])
